/////////////////////////////////////////////////////////////////////
//
// ZTFSTreeModel.java :
//
//  Data model of the main frame's file tree view.
//
/////////////////////////////////////////////////////////////////////
//
// $Revision: 4 $
// $Date: 7/11/03 1:42a $
//
/////////////////////////////////////////////////////////////////////

package com.madsci.zoot;

import java.io.*;
import java.util.Vector;
import javax.swing.event.*;
import javax.swing.tree.*;

/**
 * Data model of the main frame's file tree view.
 */
class ZTFSTreeModel implements TreeModel
{
	//
	// Contains an in-memory representation of a File object
	//
	class PseudoFile
	{
		protected String sName = null;
		protected String sPath = null;
		protected boolean bIsDir = false;
		protected PseudoFile pfKids[] = null;

		PseudoFile(File f)
		{
			try
			{
				bIsDir = f.isDirectory();
				sName  = f.getName();
				sPath  = f.getCanonicalPath();
			}
			catch (Exception ex) { }
		}

		public boolean isLeaf() { return !bIsDir; }
		public String toString()
		{
			return sName == null ? Res.get("UNKNOWNNODE") : sName;
		}

		public String getPath() { return sPath; }
	}

	private Vector treeModelListeners = new Vector();

	private PseudoFile root;

	public ZTFSTreeModel(File fRoot)
	{
		root = new PseudoFile(fRoot);
	}

	public String toString()
	{
		return root.toString();
	}

	public String getRootPath()
	{
		return root.getPath();
	}

	public String getRelativePath(PseudoFile pf)
	{
		String sFullNodePath = pf.getPath();
		String sRootPath = root.getPath();

		if (sRootPath.charAt(sRootPath.length()-1) != File.separatorChar)
			sRootPath += File.separator;

		if (sFullNodePath.indexOf(sRootPath) != 0)
			return sFullNodePath;

		return sFullNodePath.substring(sRootPath.length());
	}

	public File getFile(Object obj)
	{
		if (!(obj instanceof PseudoFile))
			return new File("");
		return new File(((PseudoFile)obj).getPath());
	}
	/**
	 * Refresh our in-memory representation of the file system
	 */
	public Object refresh(File fRoot)
	{
		root = new PseudoFile(fRoot);
		fireTreeStructureChanged(root);
		return root;
	}

	/**
	 * The only event raised by this model is TreeStructureChanged with the
	 * root as path, i.e. the whole tree has changed.
	 */
	protected void fireTreeStructureChanged(PseudoFile oldRoot)
	{
		int len = treeModelListeners.size();
		TreeModelEvent e = new TreeModelEvent(this,
											  new Object[] {oldRoot});
		for (int i = 0; i < len; i++) {
			((TreeModelListener)treeModelListeners.elementAt(i)).
					treeStructureChanged(e);
		}
	}

	/**
	 * Adds a listener for the TreeModelEvent posted after the tree changes.
	 */
	public void addTreeModelListener(TreeModelListener l)
	{
		treeModelListeners.addElement(l);
	}

	/**
	 * Returns the child of parent at index index in the parent's child array.
	 */
	public Object getChild(Object parent, int index)
	{
		if (index < 0)
			return null;

		File fKids[] = null;
		PseudoFile daddy = (PseudoFile)parent;
		if (daddy.bIsDir)
			fKids = new File(daddy.getPath()).listFiles();

		if (fKids == null || index >= fKids.length)
			return null;

		return new PseudoFile(fKids[index]);
	}

	/**
	 * Returns the number of children of parent.
	 */
	public int getChildCount(Object parent)
	{
		File fKids[] = null;
		PseudoFile daddy = (PseudoFile)parent;
		if (daddy.bIsDir)
			fKids = new File(daddy.getPath()).listFiles();

		return fKids == null ? 0 : fKids.length;
	}

	/**
	 * Returns the index of child in parent.
	 */
	public int getIndexOfChild(Object parent, Object child)
	{
		PseudoFile fParent = (PseudoFile)parent;
		PseudoFile fKid    = (PseudoFile)child;

		File fKids[] = new File(fParent.getPath()).listFiles();
		for (int i = 0; i < fKids.length; i++)
		{
			try
			{
				if (fKids[i].getCanonicalPath().equals(fKid.getPath()))
					return i;
			}
			catch (IOException e) { }
		}
		return -1;
	}

	/**
	 * Returns the root of the tree.
	 */
	public Object getRoot()
	{
		return root;
	}

	/**
	 * Returns true if node is a leaf.
	 */
	public boolean isLeaf(Object node)
	{
		return !((PseudoFile)node).bIsDir;
	}

	/**
	 * Removes a listener previously added with addTreeModelListener().
	 */
	public void removeTreeModelListener(TreeModelListener l)
	{
		treeModelListeners.removeElement(l);
	}

	/**
	 * Messaged when the user has altered the value for the item
	 * identified by path to newValue.  Not used by this model.
	 */
	public void valueForPathChanged(TreePath path, Object newValue)
	{
		System.out.println("*** valueForPathChanged : "
						   + path + " --> " + newValue);
	}
}

/********************************************************************
 *
 * $Log: /Zoot/src/com/madsci/zoot/ZTFSTreeModel.java $
 *
 * 4     7/11/03 1:42a Fpoulin
 * Optimized imports
 *
 * 3     7/08/03 4:40a Fpoulin
 * Added header and log comments.
 *
********************************************************************/
