/////////////////////////////////////////////////////////////////////
//
// ZTImage.java :
//
//  A ZT Image.
//
/////////////////////////////////////////////////////////////////////
//
// $Revision: 11 $
// $Date: 11/11/09 3:40p $
//
/////////////////////////////////////////////////////////////////////

package com.madsci.zoot;

import java.awt.*;
import java.awt.image.*;
import java.io.*;
import java.nio.*;
import java.nio.channels.FileChannel;
import java.util.*;
import javax.imageio.ImageIO;

/**
 * A ZT Image.
 */
public class ZTImage
{
	public int nFrameRate = 0x3E8;
	public int nNumFrames = 0;
	public String sFileName = null;
	public String sRootDir = null;

	Vector vecFrames = new Vector();

	ZTPal ztpal = new ZTPal();

	ZTImage(int nFrames, int nFR)
	{
		nNumFrames = nFrames;
		nFrameRate = nFR;
	}
	ZTImage(String sRoot, String sName, int nFR)
	{
		sFileName = sName;
		sRootDir = sRoot;

		nFrameRate = nFR;
		nNumFrames = 0;
	}

	public boolean isValidImage()
	{
		boolean bRet = false;

		if (sFileName == null || sFileName.length() == 0 ||
		    sRootDir == null  || sRootDir.length()  == 0)
			return false;

		File f = new File(sRootDir + File.separator + sFileName);

		// Jay added the check that the file is at least 14 bytes
		// long, which allows for the framerate, palette file name
		// length, 1 character palette file name, null, and number
		// of frames.  The file should be greater than that,
		// but that is a minimum.

		if (!f.exists() || !f.isFile() || !f.canRead()
		    || f.length() < 14)
			return false;

		RandomAccessFile raf = null;
		FileChannel rChannel = null;
		ByteBuffer bb = null;

		//
		// Create a read-only memory-mapped file
		//
		try
		{
			raf = new RandomAccessFile(f, "r");
			rChannel = raf.getChannel();
			bb = rChannel.map(FileChannel.MapMode.READ_ONLY, 0, (int)rChannel.size());
			bb.order(ByteOrder.LITTLE_ENDIAN);

			//
			// The first value is either a "Magic" number, or the framerate -- there
			// are two different file formats.
			//
			nFrameRate = bb.getInt();

			if (nFrameRate == 0x5a544146) // "FATZ" -- different file format!
			{
				//
				// Here's a bunch of bytes that I don't know what they do...
				//
				bb.position(9);

				//
				// Now here's the framerate. The rest of the file has the same format
				// in both cases.
				//
				nFrameRate = bb.getInt();
			}

			//
			// Length of the palette file name, including null
			//
			int nStrLen = bb.getInt();

			// Jay added the following check that the palette
			// file name length is at least 2 bytes long (1 byte
			// file name followed by null, although it should
			// always be greater) and not greater than the file
			// length (excluding the bytes for the frame rate,
			// palette file name length, and number of frames).
			// If it is outside this range, the file is not a ZT
			// image and might be a text file without a ".txt"
			// extension, such as a MANIFEST file.

			if (nStrLen < f.length() - 11 && nStrLen > 1)
			{ // Palette file name

			  byte bbPalName[] = new byte[nStrLen];
			  bb.get(bbPalName, 0, nStrLen);
			  String sPalName = new String(bbPalName, 0, nStrLen-1);

			  // See if we can find the palette

			  File fPal = new File(sRootDir + File.separator + sPalName);

			  if (fPal.exists() && f.isFile())
			  { bRet = true; // Good enough
			  }
			}
		}
		catch (Exception e) {}
		finally
		{
			bb = null;
			try
			{
				if (raf != null)
					raf.close();
				if (rChannel != null)
					rChannel.close();
			}
			catch(IOException ioe) {}
			finally { rChannel = null; raf = null; }
		}

		return bRet;
	}

	public boolean load()
		throws IOException, ConversionException
	{
		if (sFileName == null || sFileName.length() == 0 ||
		    sRootDir == null  || sRootDir.length()  == 0)
			throw new ConversionException(Res.get("ERR_IMGNAMENOTSET"));

		File f = new File(sRootDir + File.separator + sFileName);

		if (!f.exists() || !f.isFile() || !f.canRead())
		{
			Object args[] = { f.getCanonicalPath() };
			throw new IOException(Res.format("ERR_FILEACCESS", args));
		}

		// Jay added the following 4 lines to check that the file
		// is at least 14 bytes long, which allows for the framerate,
		// palette file name length, 1 character palette file name,
		// null, and number of frames.  The file should be greater
		// than that, but that is a minimum.  If the file is not
		// at least 14 bytes long, an exception is thrown indicating
		// it is not a ZT image file.

		if (f.length() < 14)
		{ Object args[] = { f.getCanonicalPath() };
		  throw new IOException(Res.format("ERR_NOTZTIMAGE", args));
		}

		RandomAccessFile raf = null;
		FileChannel rChannel = null;
		ByteBuffer bb = null;

		try
		{
			//
			// Create a read-only memory-mapped file
			//
			raf = new RandomAccessFile(f, "r");
			rChannel = raf.getChannel();
			bb = rChannel.map(FileChannel.MapMode.READ_ONLY, 0, (int)rChannel.size());
			bb.order(ByteOrder.LITTLE_ENDIAN);

			//
			// The first value is either a "Magic" number, or the framerate -- there
			// are two different file formats.
			//
			nFrameRate = bb.getInt();

			if (nFrameRate == 0x5a544146) // "FATZ" -- different file format!
			{
				//
				// Here's a bunch of bytes that I don't know what they do...
				//
				bb.position(9);

				//
				// Now here's the framerate. The rest of the file has the same format
				// in both cases.
				//
				nFrameRate = bb.getInt();
			}

			//
			// Length of the palette file name, including null
			//
			int nStrLen = bb.getInt();

			// Jay added the following 4 lines to check that
			// the palette file name length is at least 2 bytes
			// long (1 byte file name followed by null,
			// although it should always be greater)
			// and not greater than the file length
			// (excluding the bytes for the frame rate,
			// palette file name length, and number of frames).
			// If it is outside this range, the file is not a ZT
			// image and might be a text file without a ".txt"
			// extension, such as a MANIFEST file.  If the palette
			// file name length is not in the range, an exception
			// is thrown indicating it is not a ZT image file.

			if (nStrLen > f.length() - 12 || nStrLen < 2)
			{ Object args[] = { f.getCanonicalPath() };
			  throw new IOException(Res.format("ERR_NOTZTIMAGE", args));
			}

			//
			// Palette file name
			//
			byte bbPalName[] = new byte[nStrLen];
			bb.get(bbPalName, 0, nStrLen);
			String sPalName = new String(bbPalName, 0, nStrLen-1);

			//
			// Load the palette -- we're gonna need it to load the frames.
			//
			File fPal = new File(sRootDir + File.separator + sPalName);

			if (fPal.exists() && f.isFile())
			{
				ztpal.load(sRootDir, sPalName);
			}
			else
			{
				Object args[] = { fPal.getCanonicalPath() };
				throw new IOException(Res.format("ERR_NOPALFILE", args));
			}

			//
			// # of frames
			//
			nNumFrames = bb.getInt();

			//
			// Load each frame
			//
			for (int i = 0; i < nNumFrames; i++)
			{
				ZTFrame frame = new ZTFrame((short)0, (short)0);
				frame.loadFromFile(bb);
				vecFrames.add(frame);
			}

			//
			// TODO: deleteme
			//
//			loadbogus(bb);
		}
		finally
		{
			bb = null;
			if (rChannel != null)
			{
				rChannel.close();
				rChannel = null;
			}
			if (raf != null)
			{
				raf.close();
				raf = null;
			}
		}
		return true;
	}
	class PixelSet
	{
		int yNumTrans = 0;
		Vector vecColorPixels = new Vector();
	};

	class Line
	{
		Vector vecPixelSets = new Vector();
	};

	public boolean loadbogus(ByteBuffer bb)
		throws IOException
	{
System.out.println(""+0x45bc);
		ztpal.setTransparentColor(0xFFFF00FF);
		//
		// Start out clean
		//
		int nWidth;
		int nHeight;
		int nOffsetX;
		int nOffsetY;
		nWidth = nHeight = 0;
		nOffsetX = nOffsetY = 0;
		int funk = 0;

		Vector vecBitmap = new Vector();
		vecBitmap.clear();

		try
		{
			//
			// Remember our current position
			//
			int nStartOfPixelBlock = bb.position();

			//
			// Height and width of bitmap, X and Y offsets
			//
			funk = bb.getInt();
			nHeight = bb.getShort();
			nWidth = bb.getShort();
			nOffsetY = bb.getShort();
			nOffsetX = bb.getShort();

			//
			// No idea what this value means...
			//
			short nUnused = bb.getShort();

			//
			// Read each line of the frame
			//
			for (int nLine = 0; nLine < nHeight; nLine++)
			{
				Line l = new Line();
				vecBitmap.add(l);

				//
				// Get the number of pixel sets in this line
				//
				byte yNumSets = bb.get();

				//
				// Read in and interpret each pixel set
				//
				for (byte ySet = 0; ySet < yNumSets; ySet++)
				{
					PixelSet ps = new PixelSet();
					l.vecPixelSets.add(ps);

					//
					// Number of transparent pixels in this set
					//
					ps.yNumTrans = (bb.get() & 0xFF);

					//
					// Number of color pixels in this set
					//
					int nNumColorPixels = (bb.get() & 0xFF);

					//
					// Add the color pixels to the set
					//
					for (int i = 0; i < nNumColorPixels; i++)
					{
						byte yPixel = bb.get();
						ps.vecColorPixels.add(new Short((short)(yPixel&0xFF)));
					}
				}
			}

			//
			// Grab the transparent color now -- we're gonna use it a lot.
			// Force an alpha of 0 in case this is supported by the file format.
			//
			int rgbTrans = 0;
			try
			{
				rgbTrans = ZTPal.setA(ztpal.getTransparentColorARGB(), 0x00);
			}
			catch (ConversionException e)
			{
				e.printStackTrace();  //To change body of catch statement use Options | File Templates.
			}

//		BufferedImage bi = new BufferedImage(nWidth, nHeight, nImageType);
			BufferedImage bi = new BufferedImage(nWidth, nHeight, BufferedImage.TYPE_INT_ARGB);

			short nY = 0;
			Iterator itLines = vecBitmap.iterator();
			while (itLines.hasNext())
			{
				Line l = (Line)itLines.next();
				Vector vecBits = decodeLine(l);

				short nX = 0;
				Iterator itShort = vecBits.iterator();
				while (itShort.hasNext())
				{
					short s = ((Short)itShort.next()).shortValue();
					int rgb = 0;
					if (s == (short)-1)
					{
						//
						// Transparent pixel.
						//
						rgb = rgbTrans;
					}
					else
					{
						try
						{
							rgb = ztpal.getARGBColor(s);
						}
						catch (ConversionException e)
						{
							e.printStackTrace();  //To change body of catch statement use Options | File Templates.
						}
					}
// TODO: Optimize this by saving one line at a time?
					bi.setRGB(nX, nY, rgb);
					nX++;
				}
				nY++;
			}
			System.out.println(""+bb.position());
//			ImageIO.write(bi, "png", new File("c:\\XXX.png"));;

		}
		catch (BufferUnderflowException ex)
		{
			Object args[] = { new Integer(bb.position()) };
			throw new IOException(Res.format("ERR_ENDOFFRAME", args));
		}

		return true;
	}

	public Vector decodeLine(Line l)
	{
		Vector vecBits = new Vector();

		//
		// Go through all the PixelSets for this line
		//
		Iterator it = l.vecPixelSets.iterator();

		while (it.hasNext())
		{
			PixelSet ps = (PixelSet)it.next();
			//
			// Leading trans pixels
			//
			for (int i = 0; i < ps.yNumTrans; i++)
				vecBits.add(new Short((short)-1));

			//
			// Color pixels
			//
			Iterator itPix = ps.vecColorPixels.iterator();
			while (itPix.hasNext())
			{
				Short b = (Short)itPix.next();
				vecBits.add(b);
			}
		}
		return vecBits;
	}

	public boolean save(String sRoot, String sName)
		throws IOException, ConversionException
	{
		if (sName == null || sName.length() == 0 ||
		    sRoot == null || sRoot.length() == 0)
			throw new ConversionException(Res.get("ERR_IMGNAMENOTSET"));

		File f = new File(sRoot + File.separator + sName);

		// The following 4 lines were added by Jay to make sure
		// there are permissions to write the file.

		if (f.exists() && !f.canWrite())
		{ Object args[] = { f.getCanonicalPath() };
		  throw new IOException(Res.format("ERR_ROFILE", args));
		}

		sRootDir = sRoot;
		sFileName = sName;

		String sPalName = sName + ".pal";

		//
		// We need to force the file sep char to be "/" because
		// ZT chokes on "\" for file names in image files.
		//
		sPalName = sPalName.replaceAll("\\\\", "/");

		// Jay added the following 2 lines to have Java do the best it
		// can at getting rid of anything no longer used.  This gives
		// more room in the heap space if multiple, very large
		// animations are worked on consecutively.  Also, without it,
		// removing prior file references was not actually taking
		// place in some cases because Java thought those files still
		// had active memory maps against them when they actually
		// didn't and Java was not getting to the garbage collection
		// fast enough.

		System.runFinalization();
		System.gc();

		//
		// Make a ByteBuffer large enough to contain most images, because
		// calculating the file's total size ahead of time is an expensive
		// pain...
		//
		// The following line was changed by Jay to increase memory
		// allocated from 1 MiB to 16 MiB, since some user made
		// objects are large, especially those with animations.
		// Although it would be better to not "waste" memory
		// by calculating the space needed instead, this was an easier
		// change.  In addition, with memory page swapping,
		// this approach probably isn't hurting anything.

		ByteBuffer bb = ByteBuffer.allocate(0x1000000);
		bb.order(ByteOrder.LITTLE_ENDIAN);

		//
		// Animation frame rate
		//
		bb.putInt(nFrameRate);

		//
		// Length of palette name, including null
		//
		int nPalLen = sPalName.length() + 1;
		bb.putInt(nPalLen);

		//
		// Palette file name, with null
		//
		byte bbPalName[] = sPalName.getBytes();
		bb.put(bbPalName);
		bb.put((byte)0);

		//
		// Num frames
		//
		bb.putInt(vecFrames.size());

		//
		// Frames
		//
		for (int i = 0; i < vecFrames.size(); i++)
			((ZTFrame)(vecFrames.get(i))).saveToFile(bb);

		RandomAccessFile raf = null;
		FileChannel wChannel = null;

		// Jay added the following 2 lines to have Java do the best it
		// can at getting rid of anything no longer used.
		// Even with the System.gc() earlier in this routine,
		// there were still times when Java thought those files still
		// had active memory maps against them when they actually
		// didn't and Java was not getting to the garbage collection
		// fast enough.  So this allows for another attempt before
		// the truncate below is done.

		System.runFinalization();
		System.gc();

		try
		{
			//
			// Create a writable file channel
			//
			raf = new RandomAccessFile(f, "rw");
			wChannel = raf.getChannel();

			//
			// Write the ByteBuffer contents to the channel.
			// We call flip() on the BB first to truncate it to the current
			// position (otherwise all 1 MB of the BB will be written!)
			//
			bb.flip();
			wChannel.write(bb);

			// Jay added the following line so that the file
			// will be reduced in size, if appropriate,
			// to match that of the new graphics.

			wChannel.truncate(wChannel.position());
		}
		finally
		{
			//
			// Close the file
			//
			if (wChannel != null)
			{
				wChannel.force(true);
				wChannel.close();
				wChannel = null;
			}
			if (raf != null)
			{
				raf.close();
				raf = null;
			}

			bb = null;
		}

		//
		// Palette file
		//
		return ztpal.save(sRoot, sPalName);
	}

	public int  getNumFrames() { return nNumFrames; }
	public void setNumFrames(int nF) { nNumFrames = nF; }
	public int  getFrameRate() { return nFrameRate; }
	public void setFrameRate(int uFR) { nFrameRate = uFR; }

	public void clearFrames() { vecFrames.clear(); ztpal = new ZTPal(); }

	// Jay added the xOffset and yOffset parameters.

	public boolean addFrameFrom(BufferedImage image, short xOffset,
	                            short yOffset)
		throws ConversionException
	{
		short nW = (short)image.getWidth();
		short nH = (short)image.getHeight();

		//
		// Grab the image bits
		//
		int nBits[] = image.getRGB(0, 0, (int)nW, (int)nH, null, 0, (int)nW);

		//
		// Construct a new frame and fill it out
		//
		// Jay changed the following line to use xOffset and yOffset
		// instead of nW/2 and nH/2, which allows calling routines
		// to determine what to use.

		ZTFrame frame = new ZTFrame(nW, nH, xOffset, yOffset);

		//
		// Top left corner pixel is always considered transparent!
		//
		setTransparentColorARGB(image.getRGB(0,0));

		for (short y = 0; y < nH; y++)
		{
			int nColorValues[] = new int[nW];
			System.arraycopy(nBits, y*nW, nColorValues, 0, nW);
			frame.setPixelLine(y, nColorValues, ztpal);
		}

		//
		// Add the frame to the image
		//
		vecFrames.add(frame);

		return true;
	}

	public DisplayImage toBufferedImage(int nFrame)
		throws ConversionException
	{
		if (nFrame >= vecFrames.size())
		{
			Object args[] = { new Integer(nFrame), new Integer(vecFrames.size()) };
			throw new ConversionException(Res.format("ERR_FRAMEOOR", args));
		}
		return ((ZTFrame)vecFrames.get(nFrame)).writeToImage(ztpal);
	}

	public Point getOffsets(int nFrame)
	{
		if (vecFrames.size() == 0 || nFrame >= vecFrames.size())
			return new Point();
		return ((ZTFrame)vecFrames.get(nFrame)).getOffsets();
	}
	public void setOffsets(int nFrame, Point point)
	{
		if (vecFrames.size() == 0 || nFrame >= vecFrames.size())
			return;
		((ZTFrame)vecFrames.get(nFrame)).setOffsets(point);
	}
	public void setAllOffsets(Point point)
	{
		Iterator it = vecFrames.iterator();
		while (it.hasNext())
		{
			ZTFrame f = (ZTFrame)it.next();
			f.setOffsets(point);
		}
	}

	public void setTransparentColor(int  rgb)
	{
		ztpal.setTransparentColor(rgb);
	}
	public void setTransparentColorARGB(int  rgb)
	{
		ztpal.setTransparentColorARGB(rgb);
	}

	public static boolean isSupportedType(String sFile)
	{
		int nPos = sFile.lastIndexOf(".");

		if (nPos == -1)
			return false;

		String sExt = sFile.substring(nPos);

		if (sExt.compareToIgnoreCase(".png") == 0)
			return true;

		return false;
	}
}

/********************************************************************
 *
 * $Log: /Zoot/src/com/madsci/zoot/ZTImage.java $
 *
 * 11   11/11/09 3:40p Jay
 * Added checks that the palette file name length is a legitimate size,
 * added checks that there are permissions to write to the file, have Java
 * free unused resources before dealing with potentially large images,
 * increased the allocation size to allow much larger images, allow files
 * to reduce in size, and have the calling routines indicate what offsets
 * to use.
 *
 * 10    8/07/03 6:52p Fpoulin
 * Fixed horrible bug when creating a new ZT image.
 *
 * 9     8/05/03 12:25p Fpoulin
 * Fixed a bunch of file handle leaks.
 *
 * 8     7/17/03 12:53p Fpoulin
 * Added a ZTImage.isValidImage() call to boost tree expansion
 * performance.
 *
 * 7     7/11/03 1:42a Fpoulin
 * Optimized imports
 *
 * 6     7/09/03 7:24p Fpoulin
 * Always set the transparent color prior to converting a BufferedImage.
 *
 * 5     7/09/03 4:40p Fpoulin
 * Added API to clear all frames in the ZT image.
 *
 * 4     7/08/03 4:50a Fpoulin
 * Performance improvements.
 *
********************************************************************/
