/////////////////////////////////////////////////////////////////////
//
// ZTPal.java :
//
//  ZT palette management.
//
/////////////////////////////////////////////////////////////////////
//
// $Revision: 6 $
// $Date: 8/05/03 12:25p $
//
/////////////////////////////////////////////////////////////////////

package com.madsci.zoot;

import java.io.*;
import java.nio.*;
import java.nio.channels.FileChannel;

/**
 * ZT Palette Management.
 */
public class ZTPal
{
	private int rgbTransparent;
	// TODO: Managing this array is EXPENSIVE. Need to make this a LinkedHashMap or something.
	private int vecColors[] = new int[nMaxColors];
	private int nNumColors = 0;
	static private final int nMaxColors = 256;

	//
	// ZT Palettes are (annoyingly) in ABGR format
	//
	static private final int A_MASK = 0xFF000000;
	static private final int A_SHIFT = 24;
	static private final int R_MASK = 0x000000FF;
	static private final int R_SHIFT = 0;
	static private final int G_MASK = 0x0000FF00;
	static private final int G_SHIFT = 8;
	static private final int B_MASK = 0x00FF0000;
	static private final int B_SHIFT = 16;

	static public final int getA(int rgb) { return (rgb & A_MASK) >> A_SHIFT; }
	static public final int getR(int rgb) { return (rgb & R_MASK) >> R_SHIFT; }
	static public final int getG(int rgb) { return (rgb & G_MASK) >> G_SHIFT; }
	static public final int getB(int rgb) { return (rgb & B_MASK) >> B_SHIFT; }
	static public final int setA(int rgb, int val) { return (rgb & ~A_MASK) | ((val & 0xFF) << A_SHIFT); }
	static public final int setR(int rgb, int val) { return (rgb & ~R_MASK) | ((val & 0xFF) << R_SHIFT); }
	static public final int setG(int rgb, int val) { return (rgb & ~G_MASK) | ((val & 0xFF) << G_SHIFT); }
	static public final int setB(int rgb, int val) { return (rgb & ~B_MASK) | ((val & 0xFF) << B_SHIFT); }

	static public final int swapRedAndBlue(int val)
	{
		return ((val & (A_MASK | G_MASK)) | (val & R_MASK) << B_SHIFT | (val & B_MASK) >> B_SHIFT);
	}
	static public final int makeABGR(int r, int g, int b, int alpha)
	{
		return ((alpha & 0xFF) << A_SHIFT) |
			   ((r     & 0xFF) << R_SHIFT) |
			   ((g     & 0xFF) << G_SHIFT) |
			   ((b     & 0xFF) << B_SHIFT);
	}

	public ZTPal()
	{
		rgbTransparent = makeABGR(0x0, 0x0, 0x0, 0x00);
	}

	//
	// Load from/save to the given file name
	//
	public boolean load(String sRootDir, String sName)
		throws IOException
	{
		File f = new File(sRootDir + File.separator + sName);

		if (!f.exists() || !f.isFile() || !f.canRead())
		{
			Object args[] = { f.getCanonicalPath() };
			throw new IOException(Res.format("ERR_FILEACCESS", args));
		}

		//
		// Create a read-only memory-mapped file
		//
		RandomAccessFile raf = new RandomAccessFile(f, "r");
		FileChannel rChannel = raf.getChannel();
		ByteBuffer bbBuf = rChannel.map(FileChannel.MapMode.READ_ONLY, 0, (int)rChannel.size());
		bbBuf.order(ByteOrder.LITTLE_ENDIAN);
		IntBuffer ib = bbBuf.asIntBuffer();

		try
		{
			nNumColors = ib.get();

			vecColors = new int[nNumColors];

			for (int i = 0; i < nNumColors; i++)
				vecColors[i] = ib.get();
		}
		catch (BufferUnderflowException ex)
		{
			rChannel.close();
			Object args[] = { sName, new Integer(ib.position()) };
			throw new IOException(Res.format("ERR_EOPALETTE", args));
		}

		rChannel.close();
		raf.close();
		bbBuf = null;
		ib = null;
		rChannel = null;
		raf = null;
		return true;
	}

	public boolean save(String sRootDir, String sName)
		throws IOException
	{
		File f = new File(sRootDir + File.separator + sName);

		if (f.exists() && !f.canWrite())
		{
			Object args[] = { f.getCanonicalPath() };
			throw new IOException(Res.format("ERR_ROPALETTE", args));
		}

		RandomAccessFile raf = null;
		FileChannel rwChannel = null;
		ByteBuffer bbBuf = null;
		IntBuffer ibBuf = null;

		try
		{
			//
			// Create a read-write memory-mapped file
			//
			f.delete();
			raf = new RandomAccessFile(f, "rw");
			rwChannel = raf.getChannel();

			//
			// Calculate the overall size of the file: 4 bytes for every palette
			// entry, plus an int at the beginning for the palette size.
			//
			int nPalSize = 256;
//	    	int nPalSize = getA(rgbTransparent) == 0x0 ? 255 : 256;
			int nFileSize = (nPalSize*4) + 4;

			bbBuf = rwChannel.map(FileChannel.MapMode.READ_WRITE, 0, nFileSize);
			bbBuf.order(ByteOrder.LITTLE_ENDIAN);
			ibBuf = bbBuf.asIntBuffer();

			ibBuf.put(nPalSize);

			ibBuf.put(vecColors, 0, nNumColors);

			//
			// Fill out the rest of the palette with empty values
			//
			for (int i = nNumColors; i < nPalSize; i++)
				ibBuf.put(0xFF000000);
		}
		finally
		{
			if (rwChannel != null)
			{
				rwChannel.force(true);
				rwChannel.close();
			}
			if (raf != null)
				raf.close();
			bbBuf = null;
			ibBuf = null;
			rwChannel = null;
			raf = null;
		}

		return true;
	}

	public int getTransparentColor()
		throws ConversionException
	{
		//
		// Whenever the transparent color is set, we force the
		// alpha value to be 0xFF. So when it's 0 we know the
		// transparent color hasn't been set yet...
		//
		if (getA(rgbTransparent) == 0x00)
			throw new ConversionException(Res.get("ERR_TRANSNOTSET"));

		return rgbTransparent;
	}
	public int getTransparentColorARGB()
		throws ConversionException
	{
		return swapRedAndBlue(getTransparentColor());
	}
	public void setTransparentColor(int rgb)
	{
		//
		// Force Alpha == 0xFF
		//
		rgbTransparent = setA(rgb, 0xFF);
	}
	public void setTransparentColorARGB(int rgb)
	{
		setTransparentColor(swapRedAndBlue(rgb));
	}

	private boolean isTransparentColor(int rgb)
		throws ConversionException
	{
		//
		// Whenever the transparent color is set, we force the
		// alpha value to be 0xFF. So when it's 0 we know the
		// transparent color hasn't been set yet...
		//
		if (getA(rgbTransparent) == 0x00)
			throw new ConversionException(Res.get("ERR_TRANSNOTSET"));

		return rgb == rgbTransparent;
	}

	public int getNumColors() { return nNumColors; }

	//
	// Add a color to the list (if it's not already there...)
	// Returns the palette index of the color
	//
	public int addARGBColor(int rgb)
		throws ConversionException
	{
		//
		// We expect colors to be in ABGR format...
		//
		rgb = swapRedAndBlue(rgb);

		rgb = setA(rgb, 0xFF);
		try
		{
			if (isTransparentColor(rgb))
				return -1;
		}
		catch (ConversionException ex)
		{
		}

		//
		// See if we already have this color
		//
		for (int i = 0; i < nNumColors; i++)
		{
			if (vecColors[i] == rgb)
				return i;
		}

		//
		// Need to add the color -- but only if there's room!
		//
		if (nNumColors >= nMaxColors)
			throw new ConversionException(Res.get("ERR_TOOMANYCOLS"));

		vecColors[nNumColors] = rgb;
		return nNumColors++;
	}

	//
	// Sets rgb to the color at the given index
	//
	public int getARGBColor(int index)
		throws ConversionException
	{
		//
		// Special case: index == -1 means transparent color
		//
		if (index == -1)
		{
			if (getA(rgbTransparent) == 0x00)
				throw new ConversionException(Res.get("ERR_TRANSNOTSET"));
			return swapRedAndBlue(rgbTransparent);
		}

		if (index > nNumColors)
		{
			Object args[] = { new Integer(index), new Integer(nNumColors) };
			throw new ConversionException(Res.format("ERR_IDXOOR", args));
		}


		return swapRedAndBlue(vecColors[index]);
	}
}

/********************************************************************
 *
 * $Log: /Zoot/src/com/madsci/zoot/ZTPal.java $
 *
 * 6     8/05/03 12:25p Fpoulin
 * Fixed a bunch of file handle leaks.
 *
 * 5     7/11/03 1:42a Fpoulin
 * Optimized imports
 *
 * 4     7/08/03 4:41a Fpoulin
 * Added header and log comments.
 *
********************************************************************/
